#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
环肽氨基酸替换工具
By James-Li
"""

import os
import re
import copy
from typing import Dict, List, Tuple, Optional
from datetime import datetime
import numpy as np

AMINO_ACID_SIDECHAINS = {
    'GLY': {
        'name': 'Glycine',
        'atoms': [],  
        'bonds': []
    },
    'ALA': {
        'name': 'Alanine',
        'atoms': [('CB', 'C.3', [0.0, 0.0, 0.0])],
        'bonds': []
    },
    'VAL': {
        'name': 'Valine',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG1', 'C.3', [1.54, 0.0, 0.0]),
            ('CG2', 'C.3', [0.0, 1.54, 0.0])
        ],
        'bonds': [('CB', 'CG1'), ('CB', 'CG2')]
    },
    'LEU': {
        'name': 'Leucine',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG', 'C.3', [1.54, 0.0, 0.0]),
            ('CD1', 'C.3', [2.08, 1.3, 0.0]),
            ('CD2', 'C.3', [2.08, -1.3, 0.0])
        ],
        'bonds': [('CB', 'CG'), ('CG', 'CD1'), ('CG', 'CD2')]
    },
    'ILE': {
        'name': 'Isoleucine',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG1', 'C.3', [1.54, 0.0, 0.0]),
            ('CG2', 'C.3', [0.0, 1.54, 0.0]),
            ('CD1', 'C.3', [2.08, -1.3, 0.0])
        ],
        'bonds': [('CB', 'CG1'), ('CB', 'CG2'), ('CG1', 'CD1')]
    },
    'MET': {
        'name': 'Methionine',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG', 'C.3', [1.54, 0.0, 0.0]),
            ('SD', 'S.3', [3.08, 0.0, 0.0]),
            ('CE', 'C.3', [4.62, 0.0, 0.0])
        ],
        'bonds': [('CB', 'CG'), ('CG', 'SD'), ('SD', 'CE')]
    },
    'PHE': {
        'name': 'Phenylalanine',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG', 'C.ar', [1.54, 0.0, 0.0]),
            ('CD1', 'C.ar', [2.28, 1.28, 0.0]),
            ('CD2', 'C.ar', [2.28, -1.28, 0.0]),
            ('CE1', 'C.ar', [3.82, 1.28, 0.0]),
            ('CE2', 'C.ar', [3.82, -1.28, 0.0]),
            ('CZ', 'C.ar', [4.56, 0.0, 0.0])
        ],
        'bonds': [('CB', 'CG'), ('CG', 'CD1'), ('CG', 'CD2'), 
                 ('CD1', 'CE1'), ('CD2', 'CE2'), ('CE1', 'CZ'), ('CE2', 'CZ')]
    },
    'TYR': {
        'name': 'Tyrosine',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG', 'C.ar', [1.54, 0.0, 0.0]),
            ('CD1', 'C.ar', [2.28, 1.28, 0.0]),
            ('CD2', 'C.ar', [2.28, -1.28, 0.0]),
            ('CE1', 'C.ar', [3.82, 1.28, 0.0]),
            ('CE2', 'C.ar', [3.82, -1.28, 0.0]),
            ('CZ', 'C.ar', [4.56, 0.0, 0.0]),
            ('OH', 'O.3', [5.90, 0.0, 0.0])
        ],
        'bonds': [('CB', 'CG'), ('CG', 'CD1'), ('CG', 'CD2'), 
                 ('CD1', 'CE1'), ('CD2', 'CE2'), ('CE1', 'CZ'), ('CE2', 'CZ'), ('CZ', 'OH')]
    },
    'TRP': {
        'name': 'Tryptophan',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG', 'C.ar', [1.54, 0.0, 0.0]),
            ('CD1', 'C.ar', [2.28, 1.28, 0.0]),
            ('CD2', 'C.ar', [2.88, -1.28, 0.0]),
            ('NE1', 'N.ar', [3.82, 1.28, 0.0]),
            ('CE2', 'C.ar', [4.22, -0.72, 0.0]),
            ('CE3', 'C.ar', [3.62, -2.56, 0.0]),
            ('CZ2', 'C.ar', [5.56, -1.28, 0.0]),
            ('CZ3', 'C.ar', [4.96, -3.12, 0.0]),
            ('CH2', 'C.ar', [6.10, -2.56, 0.0])
        ],
        'bonds': [('CB', 'CG'), ('CG', 'CD1'), ('CG', 'CD2'), ('CD1', 'NE1'),
                 ('NE1', 'CE2'), ('CD2', 'CE2'), ('CD2', 'CE3'), ('CE2', 'CZ2'),
                 ('CE3', 'CZ3'), ('CZ2', 'CH2'), ('CZ3', 'CH2')]
    },
    'SER': {
        'name': 'Serine',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('OG', 'O.3', [1.41, 0.0, 0.0])
        ],
        'bonds': [('CB', 'OG')]
    },
    'THR': {
        'name': 'Threonine',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('OG1', 'O.3', [1.41, 0.0, 0.0]),
            ('CG2', 'C.3', [0.0, 1.54, 0.0])
        ],
        'bonds': [('CB', 'OG1'), ('CB', 'CG2')]
    },
    'CYS': {
        'name': 'Cysteine',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('SG', 'S.3', [1.78, 0.0, 0.0])
        ],
        'bonds': [('CB', 'SG')]
    },
    'ASN': {
        'name': 'Asparagine',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG', 'C.2', [1.54, 0.0, 0.0]),
            ('OD1', 'O.2', [2.08, 1.21, 0.0]),
            ('ND2', 'N.am', [2.08, -1.32, 0.0])
        ],
        'bonds': [('CB', 'CG'), ('CG', 'OD1'), ('CG', 'ND2')]
    },
    'GLN': {
        'name': 'Glutamine',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG', 'C.3', [1.54, 0.0, 0.0]),
            ('CD', 'C.2', [3.08, 0.0, 0.0]),
            ('OE1', 'O.2', [3.62, 1.21, 0.0]),
            ('NE2', 'N.am', [3.62, -1.32, 0.0])
        ],
        'bonds': [('CB', 'CG'), ('CG', 'CD'), ('CD', 'OE1'), ('CD', 'NE2')]
    },
    'ASP': {
        'name': 'Aspartic acid',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG', 'C.2', [1.54, 0.0, 0.0]),
            ('OD1', 'O.2', [2.08, 1.21, 0.0]),
            ('OD2', 'O.co2', [2.08, -1.21, 0.0])
        ],
        'bonds': [('CB', 'CG'), ('CG', 'OD1'), ('CG', 'OD2')]
    },
    'GLU': {
        'name': 'Glutamic acid',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG', 'C.3', [1.54, 0.0, 0.0]),
            ('CD', 'C.2', [3.08, 0.0, 0.0]),
            ('OE1', 'O.2', [3.62, 1.21, 0.0]),
            ('OE2', 'O.co2', [3.62, -1.21, 0.0])
        ],
        'bonds': [('CB', 'CG'), ('CG', 'CD'), ('CD', 'OE1'), ('CD', 'OE2')]
    },
    'LYS': {
        'name': 'Lysine',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG', 'C.3', [1.54, 0.0, 0.0]),
            ('CD', 'C.3', [3.08, 0.0, 0.0]),
            ('CE', 'C.3', [4.62, 0.0, 0.0]),
            ('NZ', 'N.3', [6.16, 0.0, 0.0])
        ],
        'bonds': [('CB', 'CG'), ('CG', 'CD'), ('CD', 'CE'), ('CE', 'NZ')]
    },
    'ARG': {
        'name': 'Arginine',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG', 'C.3', [1.54, 0.0, 0.0]),
            ('CD', 'C.3', [3.08, 0.0, 0.0]),
            ('NE', 'N.pl3', [4.62, 0.0, 0.0]),
            ('CZ', 'C.cat', [6.16, 0.0, 0.0]),
            ('NH1', 'N.pl3', [7.16, 1.27, 0.0]),
            ('NH2', 'N.pl3', [7.16, -1.27, 0.0])
        ],
        'bonds': [('CB', 'CG'), ('CG', 'CD'), ('CD', 'NE'), ('NE', 'CZ'), ('CZ', 'NH1'), ('CZ', 'NH2')]
    },
    'HIS': {
        'name': 'Histidine',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG', 'C.ar', [1.54, 0.0, 0.0]),
            ('ND1', 'N.ar', [2.28, 1.28, 0.0]),
            ('CD2', 'C.ar', [2.88, -1.28, 0.0]),
            ('CE1', 'C.ar', [3.82, 1.28, 0.0]),
            ('NE2', 'N.ar', [4.22, -0.72, 0.0])
        ],
        'bonds': [('CB', 'CG'), ('CG', 'ND1'), ('CG', 'CD2'), ('ND1', 'CE1'), ('CE1', 'NE2'), ('NE2', 'CD2')]
    },
    'PRO': {
        'name': 'Proline',
        'atoms': [
            ('CB', 'C.3', [0.0, 0.0, 0.0]),
            ('CG', 'C.3', [1.54, 0.0, 0.0]),
            ('CD', 'C.3', [3.08, 0.0, 0.0])
        ],
        'bonds': [('CB', 'CG'), ('CG', 'CD')]  # CD will connect to backbone N
    }
}

class Mol2Reader:
    def __init__(self, filename):
        self.filename = filename
        self.atoms = []
        self.bonds = []
        self.residues = []
        self.read_mol2()
    
    def read_mol2(self):
        """读取MOL2文件"""
        with open(self.filename, 'r') as f:
            lines = f.readlines()
        
        section = None
        for line in lines:
            line = line.strip()
            if line.startswith('@<TRIPOS>'):
                section = line.replace('@<TRIPOS>', '')
                continue
            
            if section == 'ATOM' and line and not line.startswith('@'):
                parts = line.split()
                atom = {
                    'id': int(parts[0]),
                    'name': parts[1],
                    'x': float(parts[2]),
                    'y': float(parts[3]),
                    'z': float(parts[4]),
                    'type': parts[5],
                    'resid': int(parts[6]) if len(parts) > 6 else 1,
                    'resname': parts[7] if len(parts) > 7 else 'UNK',
                    'charge': float(parts[8]) if len(parts) > 8 else 0.0
                }
                self.atoms.append(atom)
            
            elif section == 'BOND' and line and not line.startswith('@'):
                parts = line.split()
                bond = {
                    'id': int(parts[0]),
                    'atom1': int(parts[1]),
                    'atom2': int(parts[2]),
                    'type': parts[3] if len(parts) > 3 else '1'
                }
                self.bonds.append(bond)
    
    def get_residues(self):
        """获取所有残基信息"""
        residues = {}
        for atom in self.atoms:
            resid = atom['resid']
            if resid not in residues:
                residues[resid] = {
                    'resname': atom['resname'],
                    'atoms': []
                }
            residues[resid]['atoms'].append(atom)
        return residues

class CyclicPeptideMutator:
    def __init__(self, mol2_file):
        self.reader = Mol2Reader(mol2_file)
        self.atoms = copy.deepcopy(self.reader.atoms)
        self.bonds = copy.deepcopy(self.reader.bonds)
        self.residues = self.reader.get_residues()
    
    def find_residue_to_replace(self, target_resname=None, target_resid=None):
        """找到要替换的残基"""
        if target_resid:
            return target_resid if target_resid in self.residues else None
        
        print("检测到以下残基：")
        for resid, res_info in self.residues.items():
            print(f"  残基 {resid}: {res_info['resname']} ({len(res_info['atoms'])} 个原子)")
        
        while True:
            try:
                choice = input("\n请输入要替换的残基ID: ")
                resid = int(choice)
                if resid in self.residues:
                    return resid
                else:
                    print(f"错误：残基 {resid} 不存在")
            except ValueError:
                print("错误：请输入有效的残基ID")
    
    def get_backbone_atoms(self, resid):
        """获取残基的主链原子（N, CA, C）"""
        backbone = {'N': None, 'CA': None, 'C': None}
        for atom in self.residues[resid]['atoms']:
            atom_name = atom['name']
            if atom_name in backbone:
                backbone[atom_name] = atom
        return backbone
    
    def remove_sidechain(self, resid):
        """移除残基的侧链原子"""
        backbone_names = {'N', 'CA', 'C', 'O', 'H', 'HA', 'H1', 'H2', 'H3', 'OXT'}
        atoms_to_remove = []
        bonds_to_remove = []
        
        # 找到要删除的侧链原子
        for atom in self.residues[resid]['atoms']:
            if atom['name'] not in backbone_names:
                atoms_to_remove.append(atom['id'])
        
        # 找到涉及这些原子的键
        for bond in self.bonds:
            if bond['atom1'] in atoms_to_remove or bond['atom2'] in atoms_to_remove:
                bonds_to_remove.append(bond)
        
        # 删除原子和键
        self.atoms = [atom for atom in self.atoms if atom['id'] not in atoms_to_remove]
        self.bonds = [bond for bond in self.bonds if bond not in bonds_to_remove]
        
        return atoms_to_remove
    
    def add_sidechain(self, resid, aa_code):
        """添加新的侧链"""
        if aa_code not in AMINO_ACID_SIDECHAINS:
            raise ValueError(f"Unknown amino acid: {aa_code}")
        
        sidechain = AMINO_ACID_SIDECHAINS[aa_code]
        if not sidechain['atoms']:  # 甘氨酸没有侧链
            return
        
        # 获取CA原子
        ca_atom = None
        for atom in self.atoms:
            if atom['resid'] == resid and atom['name'] == 'CA':
                ca_atom = atom
                break
        
        if not ca_atom:
            raise ValueError(f"Cannot find CA atom for residue {resid}")
        
        # 计算新原子的起始ID
        max_atom_id = max(atom['id'] for atom in self.atoms)
        
        # 添加侧链原子
        sidechain_atoms = {}
        for atom_name, atom_type, coords in sidechain['atoms']:
            max_atom_id += 1
            new_atom = {
                'id': max_atom_id,
                'name': atom_name,
                'x': ca_atom['x'] + coords[0],
                'y': ca_atom['y'] + coords[1],
                'z': ca_atom['z'] + coords[2],
                'type': atom_type,
                'resid': resid,
                'resname': aa_code,
                'charge': 0.0
            }
            self.atoms.append(new_atom)
            sidechain_atoms[atom_name] = new_atom
        
        # 添加CA-CB键
        if 'CB' in sidechain_atoms:
            max_bond_id = max(bond['id'] for bond in self.bonds) if self.bonds else 0
            max_bond_id += 1
            self.bonds.append({
                'id': max_bond_id,
                'atom1': ca_atom['id'],
                'atom2': sidechain_atoms['CB']['id'],
                'type': '1'
            })
        
        # 添加侧链内部的键
        for bond_def in sidechain['bonds']:
            atom1_name, atom2_name = bond_def
            if atom1_name in sidechain_atoms and atom2_name in sidechain_atoms:
                max_bond_id += 1
                self.bonds.append({
                    'id': max_bond_id,
                    'atom1': sidechain_atoms[atom1_name]['id'],
                    'atom2': sidechain_atoms[atom2_name]['id'],
                    'type': '1'
                })
        
        # 更新残基名称
        for atom in self.atoms:
            if atom['resid'] == resid:
                atom['resname'] = aa_code
    
    def mutate_residue(self, resid, new_aa):
        """突变指定残基"""
        # 移除原有侧链
        removed_atoms = self.remove_sidechain(resid)
        print(f"移除了 {len(removed_atoms)} 个侧链原子")
        
        # 添加新侧链
        self.add_sidechain(resid, new_aa)
        sidechain_info = AMINO_ACID_SIDECHAINS[new_aa]
        print(f"添加了 {len(sidechain_info['atoms'])} 个 {sidechain_info['name']} 侧链原子")
    
    def write_mol2(self, filename):
        """写入MOL2文件"""
        with open(filename, 'w') as f:
            # 分子信息
            f.write("@<TRIPOS>MOLECULE\n")
            f.write("Mutated_Cyclic_Peptide\n")
            f.write(f"{len(self.atoms)} {len(self.bonds)} 0 0 0\n")
            f.write("SMALL\n")
            f.write("GASTEIGER\n\n")
            
            # 原子信息
            f.write("@<TRIPOS>ATOM\n")
            for atom in self.atoms:
                f.write(f"{atom['id']:>7} {atom['name']:<8} "
                       f"{atom['x']:>10.4f} {atom['y']:>10.4f} {atom['z']:>10.4f} "
                       f"{atom['type']:<8} {atom['resid']:>3} {atom['resname']:>3} "
                       f"{atom['charge']:>8.4f}\n")
            
            # 键信息
            if self.bonds:
                f.write("\n@<TRIPOS>BOND\n")
                for bond in self.bonds:
                    f.write(f"{bond['id']:>5} {bond['atom1']:>5} {bond['atom2']:>5} {bond['type']:>5}\n")

def main():
    print("=" * 60)
    print("环肽氨基酸替换工具")
    print("=" * 60)
    
    # 获取输入文件
    mol2_file = input("请输入MOL2文件路径: ").strip()
    if not os.path.exists(mol2_file):
        print(f"错误：文件 {mol2_file} 不存在")
        return
    
    # 读取分子
    try:
        mutator = CyclicPeptideMutator(mol2_file)
        print(f"成功读取文件：{mol2_file}")
        print(f"包含 {len(mutator.atoms)} 个原子，{len(mutator.bonds)} 个键")
    except Exception as e:
        print(f"读取文件时出错：{e}")
        return
    
    # 选择要替换的残基
    target_resid = mutator.find_residue_to_replace()
    if not target_resid:
        print("未找到要替换的残基")
        return
    
    original_resname = mutator.residues[target_resid]['resname']
    print(f"将替换残基 {target_resid} ({original_resname})")
    
    # 创建输出目录
    timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
    output_dir = f"mutations_{timestamp}"
    os.makedirs(output_dir, exist_ok=True)
    
    # 生成所有突变体
    print(f"\n开始生成18种氨基酸突变体...")
    
    success_count = 0
    for aa_code in AMINO_ACID_SIDECHAINS.keys():
        try:
            # 重新创建突变器（恢复原始状态）
            current_mutator = CyclicPeptideMutator(mol2_file)
            
            # 进行突变
            current_mutator.mutate_residue(target_resid, aa_code)
            
            # 写入文件
            output_file = os.path.join(output_dir, f"mutation_{aa_code}.mol2")
            current_mutator.write_mol2(output_file)
            
            aa_name = AMINO_ACID_SIDECHAINS[aa_code]['name']
            print(f"✓ 生成 {aa_code} ({aa_name}) -> {output_file}")
            success_count += 1
            
        except Exception as e:
            print(f"✗ 生成 {aa_code} 时出错: {e}")
    
    print(f"\n突变完成！")
    print(f"成功生成 {success_count}/18 个突变体")
    print(f"输出目录: {output_dir}")
    
    # 生成总结文件
    summary_file = os.path.join(output_dir, "mutation_summary.txt")
    with open(summary_file, 'w') as f:
        f.write("环肽氨基酸突变总结\n")
        f.write("=" * 40 + "\n\n")
        f.write(f"原始文件: {mol2_file}\n")
        f.write(f"突变位点: 残基 {target_resid} ({original_resname})\n")
        f.write(f"生成时间: {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}\n\n")
        f.write("生成的突变体:\n")
        for aa_code, info in AMINO_ACID_SIDECHAINS.items():
            f.write(f"  {aa_code}: {info['name']}\n")
    
    print(f"突变总结已保存到: {summary_file}")

if __name__ == "__main__":
    main()